/***************************************************************
ATtiny24, taktowanie zewntrznym kwarcem 16MHz + wczony BOD
Program do wielofunkcyjnego przekanika czasowego dla AVT
Micha Kurzela @ 2018
*****************************************************************/

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <avr/eeprom.h> 
#include <avr/wdt.h> 
#include <inttypes.h>

//deklaracje wyj sterujcych 74HC595 od wywietlacza LEDowego
#define LED_STCP_1 PORTA|=0b00001000
#define LED_SHCP_1 PORTA|=0b00010000
#define LED_DS_1 PORTA|=0b01000000
#define LED_STCP_0 PORTA&=0b11110111
#define LED_SHCP_0 PORTA&=0b11101111
#define LED_DS_0 PORTA&=0b10111111

//wejscia przyciskw
#define SW_H bit_is_clear(PINA,PINA0)
#define SW_M bit_is_clear(PINA,PINA1)
#define SW_S bit_is_clear(PINA,PINA2)
#define SW_H_1 bit_is_set(PINA,PINA0)
#define SW_M_1 bit_is_set(PINA,PINA1)
#define SW_S_1 bit_is_set(PINA,PINA2)

//wejcie wyzwalania
#define TRIG_1 bit_is_set(PINB,PINB2)
#define TRIG_0 bit_is_clear(PINB,PINB2)

//progi decydujce o trybie pracy
#define PROG1 100
#define PROG2 700
#define PROG3 980

//opnienie po reakcji na trzymanie przycisku [ms]
#define DELAY_1 500
#define DELAY_2 250
#define DELAY_3 100

#define IL_1 4 //ilo odebranych sygnaw z przycisku, po ktrej opnienie staje si 2x mniejsze

#define IL_2 8 //ilo odebranych sygnaw z przycisku, po ktrej opnienie staje si skrajnie mae

#define DELAY_ZAPIS 500 //tyle wywoa Timer0 trzeba, aby wykonac zapis do EEPROMu

//---------------------------------------------------------------
// Deklaracje zmiennych                                    
//---------------------------------------------------------------

//zmienne dotyczce wywietlaczy
volatile uint8_t cyfra[6] = {10, 10, 10, 10, 10, 10}; //cyfry do wystawienia na wywietlacz
volatile uint8_t cyfra_akt = 0; //aktualnie obsugiwana cyfra, od 0 do 5 (od lewej do prawej)
volatile uint8_t bajt_hc; //bajt do wysania na HC595
const uint8_t wysw[12] = { //dane o cyfrach do wysania na wywietlacz
0b11000000, //0
0b11111001, //1
0b10100100, //2
0b10110000, //3
0b10011001, //4
0b10010010, //5
0b10000010, //6
0b11111000, //7
0b10000000, //8
0b10010000, //9
0b11111111, //10, wygaszona cyfra 
0b10111111}; //11, pozioma kreska

//zmienne czasowe do wywietlenia na ekraniku
volatile uint8_t h = 0; //godziny
volatile uint8_t m = 0; //minuty
volatile uint8_t s = 0; //sekundy

//zmienne czasowe ustawione przyciskami
volatile uint8_t h_ust = 0;
volatile uint8_t m_ust = 0;
volatile uint8_t s_ust = 0;

//zmienne poprzednie, do zapisu
volatile uint8_t h_poprz = 0;
volatile uint8_t m_poprz = 0;
volatile uint8_t s_poprz = 0;

//zmienne systemowe
volatile uint8_t out = 0; //jeeli 1 to zacz wyjcie
volatile uint8_t mode = 1; //tryb pracy
//1 - podtrzymanie zaczenia przekanika przez zadany czas po ustaniu sygnau wyzwalajcego
//2 - opnienie w zaczeniu po pojawieniu si sygnau wyzwalajcego
//3 - wyczenie po ustalonym czasie od pojawienia sie sygnau (nawet jeeli trwa nadal)
//4 - zaczanie bez opnienia (zwyky przekanik on/off) - na wywietlaczach sa kreski
volatile uint8_t licznik_delay = 0; //do odliczania wartoci opnienia przy trzymaniu przycisku
volatile uint8_t trig = 0; //1 - nadszed impuls wyzwalajcy
volatile uint16_t czas_do_zapisu = DELAY_ZAPIS+1; //zmienna odliczajca od ostatniej reakcji do uaktywnienia zapisu, domylnie zablokowana
volatile uint8_t kropka = 0; //1 - pokazuje kropk na ostatniej pozycji, po zapisie w EEPROM
volatile uint16_t odlicz_1s = 0; //odlicza jedn sekund na podstawie Timera0
//---------------------------------------------------------------
// Pobiera tryb pracy
//---------------------------------------------------------------
void mode_check(void){
uint16_t adc; //warto pobrana z przetwornika
ADCSRA = 0b11100111; //obsuga ADC, podzia zegara przez 128
ADMUX = 0b00000111; //ADC7: pobranie trybu pracy

_delay_ms(2); //zaczekaj na stabilizacj wskaza

adc = (ADCL | (ADCH<<8)); //odczyt z ADC

if(adc <= PROG1){mode = 1;}
if(adc > PROG1 && adc <= PROG2){mode = 2;}
if(adc > PROG2 && adc <= PROG3){mode = 3;}
if(adc > PROG3){mode = 4;}

ADCSRA = 0; //wyczenie przetwornika A/C, nie bdzie ju potrzebny
}

//---------------------------------------------------------------
// Wysya poskadany bajt na lokalny 74HC595 (LED)
//---------------------------------------------------------------
void send_led_74hc595(uint8_t led_bajt){
uint8_t led_m; //zmiena pomocnicza do wysyki szeregowej
for(led_m=(1<<7); led_m>0; led_m>>=1){
if(led_bajt & led_m){ LED_DS_1; LED_SHCP_1; LED_SHCP_0; LED_DS_0; } //nadaj 1
else { LED_SHCP_1; LED_SHCP_0; } //nadaj 0
}}

//---------------------------------------------------------------
// Przerzuca zawarto rejestrw szeregowych
//---------------------------------------------------------------
void end_send_74hc595(void){
LED_STCP_1; LED_STCP_0; //przerzu zawarto rejestru
}

//---------------------------------------------------------------
// blok odliczania czasu
//---------------------------------------------------------------
void odliczanie(void){
odlicz_1s++;
if(odlicz_1s >= 500){odlicz_1s = 0; s--;} //odlicza sekundy
if(s >= 60){s = 59; m--;} //odlicza minuty
if(m >= 60){m = 59; h--;} //odlicza godziny
if(h == 0 && m == 0 && s == 0){out = 0; trig = 0; odlicz_1s = 0;} //zakoczenie odliczania
}

//---------------------------------------------------------------
// OBSUGA WYWIETLACZY                                                
//---------------------------------------------------------------
ISR(TIM0_OVF_vect){
TCNT0 = 131;

cyfra_akt++;
if(cyfra_akt >= 6){cyfra_akt = 0;}

bajt_hc = ~(1<<(1+cyfra_akt)); //zaczana cyfra
if(out){bajt_hc &= 0b01111111;}
send_led_74hc595(bajt_hc);

bajt_hc = wysw[cyfra[cyfra_akt]];
if(kropka && cyfra_akt == 5){bajt_hc &= 0b01111111;} //poka kropk na ostatniej pozycji
send_led_74hc595(bajt_hc);

end_send_74hc595(); //zakoczenie transmisji

if(czas_do_zapisu < DELAY_ZAPIS){czas_do_zapisu++;} //inkrementuj czas potrzebny do zapisu w EEPROM, jeeli jego odliczanie zostao uaktywniowne


if(mode == 1){ //1 - podtrzymanie zaczenia przekanika przez zadany czas po ustaniu sygnau wyzwalajcego
if(TRIG_1 && trig == 0){trig = 1; out = 1;} //jeeli wykryto zbocze narastajce, zapamitaj to i uruchom wyjcie
if(TRIG_0 && trig == 1){trig = 2; out = 1;} //jeeli impuls wyzwalajcy usta, rozpocznij odliczanie
if(trig == 2){odliczanie();}}

if(mode == 2){ //2 - opnienie w zaczeniu po pojawieniu si sygnau wyzwalajcego
if(TRIG_1 && trig == 0){trig = 1;} //jeeli wykryto zbocze narastajce, zapamitaj to
if(TRIG_0 && trig == 1){trig = 2; out = 1;} //jeeli impuls wyzwalajcy usta, rozpocznij odliczanie i uruchom wyjcie
if(trig == 2){odliczanie();}}

if(mode == 3){ //3 - wyczenie po ustalonym czasie od pojawienia si sygnau (nawet jeeli trwa nadal)
if(TRIG_1 && trig == 0){trig = 1; out = 1;} //jeeli wykryto zbocze narastajce, zapamitaj to i uruchom wyjcie
if(trig == 1){odliczanie();}}

if(mode == 4){ //4 - zaczanie bez opnienia (zwyky przekanik on/off) - na wywietlaczach s kreski
if(TRIG_1){out = 1;}
if(TRIG_0){out = 0;}
}

}

//---------------------------------------------------------------
// podzielenie na cyfry                                            
//---------------------------------------------------------------
void dziel_cyfry(void){
if(mode < 4){ //wywietlenie czasu
cyfra[0] = h/10;
cyfra[1] = h - (cyfra[0] * 10);
cyfra[2] = m/10;
cyfra[3] = m - (cyfra[2] * 10);
cyfra[4] = s/10;
cyfra[5] = s - (cyfra[4] * 10);
} else { //wywietlenie kresek w trybie 4
cyfra[0] = 11;
cyfra[1] = 11;
cyfra[2] = 11;
cyfra[3] = 11;
cyfra[4] = 11;
cyfra[5] = 11;
}}

//---------------------------------------------------------------
// GLWNA FUNKCJA PROGRAMU                                                   
//---------------------------------------------------------------
int main(void){

// opisy pinw
/*PORTA:
0 - switch H
1 - switch M
2 - switch S
3 - wyjcie STCP
4 - SCK (wyjcie SHCP)
5 - MISO
6 - MOSI (wyjcie DS)
7 - wejcie MODE analogowe
*/
DDRA = 0b01111000;
PORTA = 0b00000111;
/*PORTB:
0 - XTAL1
1 - XTAL2
2 - wejcie TRIG
3 - nieuywany
4 - nieuywany
5 - nieuywany
6 - nieuywany
7 - nieuywany
*/
DDRB = 0b11111000;
PORTB = 0b11111000;


/*Timer0 - do obsugi wywietlaczy
przeskaler przez 256, przerwania z czestotliwoscia 500Hz
+ obsuga jego przepenienia */
TCCR0A = 0;
TCCR0B = 0b00000100;
TIMSK0 |= 0b00000001;

sei();

mode_check(); //pobierz tryb pracy
//wywietl tryb pracy
cyfra[5] = mode;
_delay_ms(2000); //zaczekaj na przeczytanie
cyfra[5] = 10; //wyga
_delay_ms(500);

//odczyty z EEPROM
h_ust = eeprom_read_byte((const uint8_t *)2);
if(h_ust >= 100){h_ust = 0;} //jeeli odczytana warto jest poza zakresem, wyzeruj j

m_ust = eeprom_read_byte((const uint8_t *)4);
if(m_ust >= 60){m_ust = 0;}

s_ust = eeprom_read_byte((const uint8_t *)6);
if(s_ust >= 60){s_ust = 0;}

h_poprz = h_ust; //aby zablokowa automatyczny zapis do EEPROM
m_poprz = m_ust;
s_poprz = s_ust;

//wywietlenie czasu
dziel_cyfry();


while (1){

if(SW_H){h_ust++; if(h_ust >= 100){h_ust = 0;} licznik_delay++;} //ustawianie czasu przyciskami
if(SW_M){m_ust++; if(m_ust >= 60){m_ust = 0;} licznik_delay++;}
if(SW_S){s_ust++; if(s_ust >= 60){s_ust = 0;} licznik_delay++;}

if(trig == 0){ //jeeli odliczanie nie jest uruchomione, wywietlaj wartoci ustawione przyciskami
h = h_ust;
m = m_ust;
s = s_ust;
}

dziel_cyfry(); //wywietl

if(licznik_delay > 0 && licznik_delay <= IL_1){_delay_ms(DELAY_1); czas_do_zapisu = 0;}
if(licznik_delay > IL_1 && licznik_delay <= IL_2){_delay_ms(DELAY_2); czas_do_zapisu = 0;} //przyspiesza
if(licznik_delay > IL_2){_delay_ms(DELAY_3); czas_do_zapisu = 0;} //bardzo szybko

if(SW_H_1 && SW_M_1 && SW_S_1){licznik_delay = 0;} //jeeli wszystkie przyciski sa zwolnione, to wyzeruj opnienie


if(czas_do_zapisu == DELAY_ZAPIS){ //jeeli licznik odliczy - wykryj, co si pozmieniao i zapisz
kropka = 1; //zacz kropk na chwil
if(h_poprz != h_ust){ eeprom_write_byte((const uint8_t *)2, h_ust); h_poprz = h_ust; } //wykryta rnica = zapis do pamici
if(m_poprz != m_ust){ eeprom_write_byte((const uint8_t *)4, m_ust); m_poprz = m_ust; }
if(s_poprz != s_ust){ eeprom_write_byte((const uint8_t *)6, s_ust); s_poprz = s_ust; }
_delay_ms(100);
kropka = 0;
czas_do_zapisu = DELAY_ZAPIS + 1; //aby zdeaktywowa ponowne wejcie do tej funkcji
}

}
return 1;
}





   
   
